/* Copyright (c) 2002-2004 by Stanley M. Swanson.
   Distributed under the GNU General Public License, version 2 (GPL v2). */
/* vamp_midi.c contains code and utilities for generating the output
   MIDI file from tables derived from program input */

# include "globals.h"

# ifndef LINUX  /* my version of VisualC++ lacks the following */
# define srandom srand
# define random  rand
# define pid_t int
# endif

/* 052.16  file globals for rubato durations... */
int kgen, silence, pan;
int qacc=0, qphr=0, qrub=0, qsynch=0, quantur=60;
int now_met, when, now_step, met_frac;

void lowendian(int n, char *in, char *out)
{ int j;  if (qswap) { for (j=0; j<n; j++) out[n-1-j] = in[j]; }
          else { for (j=0; j<n; j++) out[j] = in[j]; } }

void intout(int i)
{ int j; union {int i; char in[4];}u;  u.i = i;
   if (nout+4>MAXOUT) 
   { printf("*********buffer full (intout) \n"); over_flows++; return; }  
  if (qswap) { for (j=0; j<4; j++) bout[nout+3-j] = u.in[j]; }
           else { for (j=0; j<4; j++) bout[nout+j] = u.in[j]; } 
 nout += 4; } 

void shout(short sh)
{ int j; union {short s; char sh[2];}u;  u.s= sh;
    if (nout+2>MAXOUT)
   { printf("*********buffer full (shout) \n"); over_flows++; return;}
 if (qswap) { for (j=0; j<2; j++) bout[nout+1-j] = u.sh[j]; }
 else { for (j=0; j<2; j++) bout[nout+j] = u.sh[j]; }
 nout += 2; }

void strout(char *str, int n)
{ int j;  if (nout+n>MAXOUT) 
        { printf("*********buffer full(strout) \n"); over_flows++; return; }
  for (j=0; j<n; j++) bout[nout+j] = str[j];
 nout += n ; }

void byteout(int v) /* positive byte output 0-127 */
{ char b;
 if (v<0 || v>127) { printf(" %d byte out of range\n",v); v = 0; }
 b = v; bout[nout] = b; nout++; }

void delta(int v) /* MIDI variable count, eventually up to 128^4 - 1 */
{ char b, ib[4];  int  i, n, iv = v;  
 if (v<0 || v>268435455) { printf(" %d delta out of range\n",v); v = 0; }
 for (i=0; i<4; i++) { n = i;  ib[i] = (iv & 0x7f); iv /= 128; if (iv==0) break; }
 for  (i=n; i>=0; i--) {  b = ib[i]; if (i>0) b = b | 128;  bout[nout] = b; nout++; }
}

void command(int cmd, int chan, int note, int veloc)
{ if (nout+3>MAXOUT) 
        { printf("*********buffer full(command) \n"); over_flows++; return; }
      bout[nout] = 16*(cmd & 0xf) | (chan & 0xf);
      bout[nout+1] = (note & 0x7f);
      bout[nout+2] = (veloc & 0x7f);  nout += 3; }

void change(int cmd, int chan, int prog)
{ if (nout+2>MAXOUT) 
        { printf("*********buffer full(change) \n"); over_flows++; return; }
      bout[nout] = 16*(cmd & 0xf) | (chan & 0xf);
      bout[nout+1] = (prog & 0x7f);  nout += 2; }

void midi_file_header(int org, int tracks, int delta_time, char *name)
{ int i,j,k,m,n;  short s;
  char  filnam[MAXIN], *mhead = "MThd", i4[4], j4[4], i2[2], j2[2];
  if (name) { fout = new_file(name,"wb"); }
  else { printf("output file: "); n = finstring(filnam,MAXIN-4,stdin);
      fout = fopen(filnam,"wb");}
  /* 046.02: Windows (VisualC++) appends 0x0d to every 0 byte if "w" */
  if (! fout) { printf("cannot open midi output file\n"); return; }
 /* MIDI file header */
      fwrite(mhead,1,4,fout); 
      i = 6;  lowendian(4,(char*)&i,i4);  fwrite(i4,1,4,fout); 
      s = org;  /* 0: single track; 1: one or more simultaneous tracks;
		   2: one or more sequentially independent single tracks */
         lowendian(2,(char*)&s,i2);  fwrite(i2,1,2,fout); 
      s = tracks;  /* number of tracks in file (==1 for format 0 file) */
          lowendian(2,(char*)&s,i2);  fwrite(i2,1,2,fout); 
      s = delta_time; /* ticks/quarter note>0, -SMPTE ticks/frame<0 */
          lowendian(2,(char*)&s,i2);  fwrite(i2,1,2,fout); 
}  /* midi_file_header()  */

void microseconds_in_beat(float bpm)
{ int u,usec; char i4;
      /* FF51 03 tttttt : set tempo, u sec/quarter note (24 bits) */
 u = usec = 6.0e7/bpm;   usec += 0x03000000;  /* add byte count to 24 bit integer */
 printf("bpm %g  microseconds in beat %12d  %12x\n",bpm,u,u);  
      delta(0); shout(0xFF51); intout(usec);     /* Q = 500000 usec */

  /* cakewalk puts tempo, key stuff into info track */
      /*   delta(0); shout(0xFF51); intout(0x0307A120);     /* Q = 500000 usec */
     /*   delta(0); shout(0xFF51); intout(0x0303D090);     /* Q = 250000 usec */
}

void midi_info(char *title)
{ int i,j,k,m,n,t=0;  short s;
  char  *text="experimental MIDI sequences",
    /*    *copyleft=
            "Copyright 2003 by Stanley M. Swanson under GPL, DSL, [cc]",  */
   *mhead = "MThd", *mtrk = "MTrk", i4[4], j4[4], i2[2], j2[2];

/* MIDI info track */ 
      /* FF01 len text : arbitrary text */
      /* FF02 len text : copyright notice */
      /* FF03 len text : sequence-track name */
      /* FF04 len text : instrument name */
      /* FF2F 0 : end of track (required) */
      /* FF58 04 nn dd cc bb : time signature nn/2**dd, 
	           cc clocks/metronome, bb ? */
      /* FF59 02 sf mi : key signature sf<0 flats, >0 sharps, ==0 key of C; 
	           mi 0 major, 1 minor */
      /* FF51 03 tttttt : set tempo, u sec/quarter note (24 bits) */

  if (! fout) return;
     nout = 0;

     if (copyleft) { delta(0); shout(0xFF02); 
               i = strlen(copyleft); delta(i); strout(copyleft,i); }

   i = strlen(commenta);  
      if (i) { delta(0); shout(0xFF01); delta(i); strout(commenta,i);  t=0; 
               if (Q_VERBOSE>4 && V_CLASS==4) printf(" commenta strlen %d\n",i); }

   i = strlen(commentb);  
      if (i) { delta(0); shout(0xFF01); delta(i); strout(commentb,i); t=0;
               if (Q_VERBOSE>4 && V_CLASS==4) printf(" commentb strlen %d\n",i); }

      delta(0);
      shout(0xFF03); i = strlen(title); delta(i); strout(title,i);
            if (Q_VERBOSE>4 && V_CLASS==4) printf(" strlen title %d\n",i);      
/* 046.18 suppress: t==0 */
      if (t) {delta(0); shout(0xFF01); 
              i = strlen(text); delta(i); strout(text,i);
              if (Q_VERBOSE>4 && V_CLASS==4) printf(" strlen text %d\n",i); }

      microseconds_in_beat(bpm);

     /*      delta(0); shout(0xFF58); delta(4); intout(0x04026060);  */
         /* 4/4, 0x60 clocks/beat, 0x60 ticks/Quarternote */
     /*  delta(0); shout(0xFF58); delta(4); intout(0x03024040);  */
         /* 3/4, 0x40 clocks/beat, 0x40 ticks/Quarternote */
 /* 046.18 
  *   delta(0); shout(0xFF58); delta(4);
  *      delta(meter_num);   
  *      k = 2; if (meter_denom==2) k=1; if(meter_denom==8) k=3;
  *      delta(k);  delta(0x60); delta(0x60);
  */    
  
      delta(0); shout(0xFF2F); delta(0);  /* end of track */
      printf(" info trk length %d %x %o \n",nout,nout,nout); 

      fwrite(mtrk,1,4,fout); 
      i = nout;  lowendian(4,(char*)&i,i4);  fwrite(i4,1,4,fout); 
      fwrite(bout,1,nout,fout); 

} /* midi_info()  */

int spawn_play(char *midifile)
{ pid_t pid=0;   if (!Q_PLAY) return 0;  /* 039.02 suppress with verbose<0 */
# ifdef LINUX      /* do not know how to do this in Windows */
   pid = fork();  /* spawn timidity to play midi file */
  if (pid<0) { printf("fork error\n"); return pid; }
  if (pid==0) { /* child process */
    /* 032.10		    execlp("timidity","timidity",midifile,NULL); */
    execlp("timidity","timidity","-A","200","-B","8,12","-Od","-s","48000",
			   midifile,NULL);
		    printf("error: return from spawn in child\n"); }
# endif
  return pid;
}

/*  030.25  dither: imprecision in note start and stop, 
   initialize in set_dither to sigma==msec (milliseconds ~ 10 to 20)
   generate randomly in dither(),
   apply on a per note-cluster basis in make_notes, with constraints
   of not starting a new cluster before previous one ends.
   chords have single start, end point, except guitar strum
*/
double tick_msec, scale_dither, trans_rand, scale_rand, sigma_rand;
int last_dither, max_dither, knt_dither, minim, dither_mode;
int pack_kount, gen_kount;

void pack_event( int *kx, int ticks, int cmd, int chan, int d1, int d2)
{  union {int i; char c[4]; } ev;  int k = *kx;
   ev.c[0] = cmd; ev.c[1] = chan; ev.c[2] = d1; ev.c[3] = d2;
   epoch[k] = ticks; event[k] = ev.i;
   pack_kount++;
   if (Q_VERBOSE>6 && V_CLASS==4) printf("%3d pack  %5d event %10x %3d\n",k,epoch[k],ev.i,ev.c[2]); 
   k++; if (k<MXNOTE) *kx = k; 
   else { over_flows++; if (k==MXNOTE) printf("*********overflow in pack_event\n");}
}  /* pack_event() */

int sort_gen_events(int *kx, int *previous, int now)
{   union {int i; char c[4]; } ev;  
    int i, j, jj=-1, k = *kx, knt=0, dur, m, n, dnow = now-max_dither-1;
    if (k<=0) return 0;  /* no events, likely double measure bar  */
    sort_ii(epoch,event,k);
    for (j=0; j<k; j++)
      { 
	ev.i = event[j];  if (Q_VERBOSE>6 && V_CLASS==4) 
	  printf("%3d epoch %5d event %10x %3d\n",j,epoch[j],ev.i,ev.c[2]);  
	if (epoch[j]>dnow) break;  knt++;  jj = j;
        dur = epoch[j] - *previous; 

   if (dur<0) printf("bad delta for event %d  epoch %d previous %d\n",
		     j,epoch[j],*previous);
          *previous = epoch[j];
   if (dur<0) continue;
	ev.i = event[j];
    gen_kount++;
	delta(dur);  command(ev.c[0], ev.c[1], ev.c[2], ev.c[3]); }
    if (Q_VERBOSE>2 && V_CLASS==4)
    printf("%4d midi events generated, now %d, initial %d final %d\n",
	   knt, now, epoch[0], epoch[k-1]);

    if (jj+1<k) { /* events remaining for times>now */
      for (m=0,n=jj+1; n<k; n++,m++) { epoch[m]=epoch[n]; event[m]=event[n]; }
      *kx = m;  return m;  /* remaining events */ }
    *kx = 0;  return 0;
} /* sort_gen_events() */

  float chord_quiet;    /* 045.30 to demphasize multiple notes -- default 8 */
 float chord_delta[8] = { 0.0, 10.6, 16.8, 21.2, 24.57, 27.36, 29.7, 31.7 }; 
void set_dither(int blur)  /* scaling parameters for dither */
{ int msec;
  chord_quiet = 0.125*Q_TESTA;  
  dither_mode = blur/100; msec = blur%100;
  pack_kount = gen_kount = 0; 
  knt_dither = 3;  sigma_rand = 0.501; /* sum random() 3 times */
  srandom(1);  /* for debug reproducibility */
  /* sigmas for counts 1..6:  0.284, 0.408, 0.501, 0.578, 0.647, 0.708 */
  scale_rand = 1.0/RAND_MAX; trans_rand = 0.5*knt_dither;
  tick_msec = 60000.0/(bpm*quarterticks);  last_dither = 0; 
  minim = quarterticks/8;  if (minim<4) minim = 4; 
    /* smallest allowed duration is 1/32 note */
     printf("dither msec/tick %6.2f ",tick_msec);
  if (msec==0) { max_dither = knt_dither = 0; 
                 printf("  zero\n"); return; }
  scale_dither = msec/(sigma_rand*tick_msec);
     printf(" sigma msec %d  scale %6.3f  mode %d\n",
	      msec,scale_dither,dither_mode);
  return;
} /* set_dither() */

 /* approximate gaussian by summing several uniform deviates */
int dither(int n)  /* (n==1) probably "melody", otherwise "chord" */
{ double sum, delta; int k;
  if (knt_dither==0) return 0;  /* no dither */
  sum = 0.0; for (k=0; k<knt_dither; k++) sum += random();
  delta = scale_dither*(scale_rand*sum - trans_rand); 
  if (delta>= 0.0) k = (delta +0.5); else k = (delta - 0.5);
  /* dither_mode:  0  random change to time on any note/note group
                   1  random change if not first beat in phrase
		   2  melody (n==1) leads synched chords
		   3  melody (n==1) lags synched chords  */
  if (dither_mode==1 && first_beat) k = 0;
  else if (dither_mode==2) { if (n>1) k = 0; else if (k>0) k = -k; }
  else if (dither_mode==3) { if (n>1) k = 0; else if (k<0) k = -k; }
  return k;
} /* dither() */

/* 052.11  velocity translation from internal log(amplitude) to
   either linear or quadratic (MS mediaplayer) */

int vel_trans[128];

int vel_trans_init(int mode)
{ int j, vel;  
  double v=1.0, frac=0.983759762; /* pow(2.0,(126/127-1)*3) */
 if (mode==2) { /* MS media player: squares velocity, XP prof */
      printf("output velocities for MS media player\n");
   for (j=127; j>0; j--) 
     { vel = 127.0*v + 0.5; vel_trans[j] = vel; v *= frac; } }
 else if (mode==1) { /* linear (old Timidity) */  frac *= frac;
      printf("linear output velocities (old Timidity)\n");
   for (j=127; j>0; j--) 
     { vel = 127.0*v + 0.5; vel_trans[j] = vel; v *= frac; } }
else { /* internal log of amplitude, 6 octaves, new timidity */
      printf("logarithmic output velocities (new Timidity)\n");
   for (j=127; j>= 0; j--) vel_trans[j] = j;  return 0; }
 vel_trans[0] = 0; return mode; 
 }  /* vel_trans_init() */

    /* accomodate slurs, ties, chopped note tails    04x.27 */
void make_notes(int n, int *tones, int *durs, int *vels, 
                       int now, int stac, int chan, int opt)
{ int j,k, now_start, now_end, vol = *vels, vel, dur=*durs, 
      bend, delta, nend, ne, be, ni, ramp;  

   if (n<1) return;  /* 04x.29  try pruning midi file */
  /* 04x.29 eventually deal with Q_VEL_CURVE, velocity_curve[] here */
/* 045.29  (N violin problem) attempt to have total chord volume approximate 'vol'  */

 if (opt==0) {  /* conventional dur-staccato, exponential decay, possible dither */
       if (vol<=0) return;  /* 045.30 try pruning midi file */
   if (vol>0 && n>1) 
   { k = n; if (k>7) k = 7; delta = chord_quiet*chord_delta[k] + 0.5;
     vel = vol - delta; if (vel<0) vel = 0; } else vel = vol;   if (vel<0) vel = 0; 
   if (vel>127) vel = 127;  /* keep in valid range [0...127] */
   vel = vel_trans[vel];
   now_start = now + dither(n); now_end = now_start + dur - stac; 
   /*  incompatible with overlapped guitar strums (rubato, -delay) 052.21 .... 
       if (now_start<=last_dither) now_start = last_dither + 1;  */
     if (now_end<=now_start + minim) { now_end = now_start + minim; }
   for (k=0; k<n; k++) pack_event(&kgen,now_start,0x9,chan,tones[k],vel);
   for (k=0; k<n; k++) pack_event(&kgen,now_end,0x8,chan,tones[k],0);
    last_dither = now_end;

 } else { /* bend slurs (n>1), chop last note */  vel = vol;
 /*  for (k=0;k<n;k++) printf(" k %d tone %d dur %d vol %d \n",k,tones[k],durs[k],vels[k]); */
   if (n>1) { bend = tones[0] - tones[n-1]; 
         /*   printf(" %d %d total bend %d\n",tones[0],tones[n-1],bend);  */
  /* check range...>15 gives headaches...*/
       if (bend<-15 || bend>15) { printf(" BEND too big\n"); error_count++; }
       if (bend != 0) pack_event(&kgen, now, 0xe, chan, 0, 64+4*bend);  
       } /* if (n>1) provisional bend from last note */
/* start last (bent) or only (n==1) note  */
    pack_event(&kgen,now,0x9,chan,tones[n-1],vel_trans[vel]); 
      nend = ne = now;
   if (n>1) { /* presume slurs or ties */
               ramp = Q_SLUR; if (ramp<0) ramp = -ramp;
       for (k=0; k<n-1; k++) { nend +=  durs[k]; vol = vels[k+1]; 
         if (opt&1) { /* Q_SLUR !=0 and style ok */
	   ne = nend - ramp;  if (Q_SLUR<0) ne += stac/2;
	   bend = 8192 + 512*(tones[k+1] - tones[n-1]);  /* target bend, NB offset */ 
           delta = ((511+ramp)/ramp)*(tones[k+1] - tones[k]); 
    /* tie has delta==0, no need for pitchbend */
	    if (delta != 0) { bend -= ramp*delta;
	    /*  printf("note %d,  %d %d bend %x delta %x  lm %x %x\n", 
		        k,tones[k],tones[k+1],bend,delta,bend&0x1ff,bend/128);  */
   /* pitch bend between notes in slur */
            for (j=0; j<ramp; j++) { bend += delta; ne++;
             pack_event(&kgen, ne, 0xe, chan, bend&0x1ff, bend/128); }
	    }  /* if (delta != 0 ) */ 
	 /* following may fight sustain decay, giving glitch ... */
	  if (vol!=vels[k])
            pack_event(&kgen,nend,0xa,chan,tones[n-1],vel_trans[vol]); }
	 }  /* if (opt&1) [when Q_SLUR && style ok] */ 
      } /* if (n>1) mess with slurs and ties */
   /* provisionally chop end of last note */
     nend += durs[n-1]; vol = vel = vels[n-1]; ne = nend - stac;
     if (qslur&2) { ramp = Q_CHOP;  ne = nend - ramp;
                      if (ramp<0) { ramp = -ramp; ne = nend - ramp - stac; }
	            delta = (vol+ramp-1)/ramp;
	            for (j=1; j<ramp; j++) { vel -= delta; ne++; 
		     if (vel>0)
                      pack_event(&kgen,ne,0xa,chan,tones[n-1],vel_trans[vel]); }
                  } /* if (opt&2) [when Q_CHOP && style ok] */
     /* finally stop note, but at low velocity after chop */
        pack_event(&kgen,ne,0x8,chan,tones[n-1],0);
  }  /* else if (opt) */
} /* make_notes() */

int start_midi_music_track(int instrum, int chan, int chords)
{  int i,k,m,n;
   char i4[4],  *mtrk="MTrk";

/* do track initialization (instruments, pan,  etc.) */

   nout = 0;    /* declare patches for channels */
   /* if (instrum==0) do not specify program, but use -I with timidity */

/* convert abc info into midi info, ramping accent each measure  */
	/* timidity ignores initial, final rests, so cheat */

   /* 031.29 need to patch this   cf. knote++
	if (vel[0]==0) vel[0] = 1; if (vel[npitch-1]==0) vel[npitch-1]=1;
   */
   gen_rep = gen_meas = 0;  next_event_epoch = 2000000000;
   now_met = gen_epoch = 1000;  /* allow ticks for negative offset in rubato */ 
      meas_mm = tictoc_mm = 0;  
	kgen = start_mm = 0;  silence = 0; 
        kacc = iacc; knote_mm = 0; last_mm = tonal_center;

    /*  delta(0); command(0xB,chan,7,mix_ch);*/  /* channel volume */   

  /* 042.20, 031.09 program change for multiple channels */ 
/* 04x.29  all of this gets merged in timidity: try one/track */
             /*    for (i=0;i<instrument_list[1];i++) */
    i = chan; 
    { k = instrument_list[3+i]-1; 
     if (k>=0) { if (k>127) k = 127; delta(0); change(0xC,i,k); 
        printf(" midi track  chan %d, instrument %d\n",i,k+1);
        if (flog) fprintf(flog,"midi track  chan %d, instrument %d\n",i,k+1); }
    }

    /* set style for instrument, do not want instr-1  */
    k = instrument_list[3+chan];  if (chords==1) k += 1000;
    m = get_style(k,0);  /* first try for instrument style. (will get default chord values 
                                     if chord instrument style not defined) */
    if (m==0) n = get_style(chan+1-1000,0); /* then check voice if default melody */

    qslur = 0;  if (Q_SLUR) qslur += 1; if (Q_CHOP) qslur +=2; 

   set_dither(Q_BLUR);  /* 030.25  use Q_BLUR as integer milliseconds */

   if (chords==1 || chords== -1) 
         {  /* chord track */ staccato = stac_ch; pan = pan_ch;}
   else  {  /* melody track  0,2 */  staccato = stac_mel; pan = pan_mel; }

      /* 04x.27 change pitch bend multiplyer to 16 */
    if (qslur) {
        delta(0); command(0xb,chan,101,0); /* rpn msb, lsb, data msb,lsb */
        delta(0); command(0xb,chan,100,0);
        delta(0); command(0xb,chan, 6,16);
        delta(0); command(0xb,chan,38,16);  
           /* LSB, may not be correct or needed */
    } /* (qslur) */

    if (pan!=64) { delta(0); command(0xB,chan,10,pan); }  /* channel pan  */

    /* debug printf("start track: chords %d instr %d stac %d\n",
       chords,k,staccato);  */

} /* start_midi_music_track() */

int gen_midi_music_track(int npi, int npf, int chan, int chords)
{  int i,k,n, tone,knt, dur, isv, vol=0,solo,potatoes, e_beat;
   int durub, delay, durn, rq, rub_beg, silent,delta_when;  
   double dq, ddd;
   int kord[MXLIST+1],inkord,kordur,voice,trans,left; 
   int slur[MXLIST+1],inslur,durs[MXLIST+1],slurdur,when_slur,vels[MXLIST+1];

/* adds commands to an existing MIDI track (mostly notes) */
/* MIDI commands */
      /* 8[0-F] nn vv : channel [0-F] off, nn = note, vv = velocity */
      /* 9[0-F] nn vv : channel [0-F] on,  nn = note, vv = velocity */
      /* A[0-F] nn aa : polyphonic aftertouch, nn = note, aa = aftertouch */
      /* B[0-F] cc vv : control/mode change (see table 3) */
      /* C[0-F] nn    : program change, nn = patch */
      /* D[0-F] aa    : aftertouch, aa = amount */
      /* E[0-F] lsb msb : pitch wheel control lsb (0-127), msb (0-127)*128 */
      /* F[0-F] ...  system commands */

   /* debug  printf("gentrack qacc %d qsynch %d qphr %d chord %d  kacc %d\n",
      qacc,qsynch,qphr,chords,kacc);  */

  knt = inkord = inslur = 0;  /* presume not inside a chord or slur */
  voice = 0; trans = trans_mel; 
  if (chords==1) { voice = 1;  trans = trans_ch; }
  solo = potatoes = mix_mel; 
     if (mix_ch<solo && mix_ch>solo/3) potatoes = mix_ch;
     if (Q_VERBOSE>2 && V_CLASS==4) 
         printf(">> chords %d chan %d first %d last %d \n",chords,chan,npi,npf); 
   for (n=npi; n<npf; n++) 
   { tone = pitch[n]; delta_when = durub = dur = tics[n];  silent = 0;
     if (tone>=0) { /* determine velocity: 052.17, 042.29 cf: get_velocities() */
          silence = staccato;
       switch (chords) { 
       case -1: vol = potatoes; break;
       case 0: 
       case 2:  /* as of 054.. can have acc or phrase for either */ 
         vol = solo; 
         if (qphr) { vol = veloc[nbeat]; if (vol==0) silent = 1;} 
         if (qacc) { isv = phrasing[kacc]%1000; if (isv>500) isv -= 1000; 
                     vol += isv; } break;
       case 1:  vol = mix_ch;   /*  chords */
         if (qphr) { vol = veloc[nbeat]; if (vol==0) silent = 1;} break;
         if (qacc) { isv = phrasing[kacc]; 
	         /* 02x.06 high digit flags: +i0000, change instrument;
		                              +s000 use staccato to shorten  */
		if (isv>9999) 
		  { int i; i = isv / 10000; isv -= i*10000;
		    chan = i - 1;  /* 031.09, 042.28 try channel change */
                    if (Q_VERBOSE && knt==0) 
                      printf(" instr change %d %d chan %d\n",i,instrument_list[3+i],chan);
		   }  /* (isv>9999) */
		if (isv>999) {  int i; i = isv / 1000; isv -= i*1000;
		                silence = i*staccato;
	        	if (silence>(3*dur)/4) silence = (3*dur)/4; }
                if (isv>500) isv -= 1000;  /* allow negative deltas 053.30 */
		vol += isv;  /* incremental volume 053.05 */
	 } /* if (qacc) */		break;
  /* old  case 2:  vol = veloc[nbeat]; if (vol==0) silent = 1; break; 054.27 */

       default:    if (knote_mm<3) printf("bad chord flag %d in gen_midi_ \n",chords);
                vol = solo;
       } /* switch (chords) */

       /*  052.22:  need to rework delta accent, default or phrased velocities */
       /*     if (qrub && qacc && (chords==0 || chords==2)) vol +=  phrasing[kacc]; */

       vol += volume_delta;  /* 045.03 _volume_ in events */
       if (n_emph>0) { e_beat = tick_sum/ticks_in_beat; vol += emphasis[e_beat]; }
       if (vol<0) vol = 0;  
             /* final rest in [chord] will cancel volume (might need vol[] )
                 probably not true anymore (053.30) */ 
       if (tone==E_REST || silent) 
	        { tone = last_mm; vol = 0;
                     /* possible Timidity synch problem on initial rest: */
                            if (knote_mm==0) vol =1; /* 031.29  initial rest */}

/*debug  printf("tone %d vol %d dur %d  kacc %d  delvol %d\n",
  tone,vol,dur,kacc,phrasing[kacc]); */

       if (inkord || inslur ) vels[knt] = vol; 

       if (inkord)   /* accumulate metric durations until E_END_CHORD */
       { if (knt==0) kordur = dur;  if (tone>0) last_mm = tone;   
         tone += trans;  if (tone<0) tone = 0; if (tone>127) tone = 127;
         kord[knt] = tone;  durs[knt] = dur; vels[knt] = vol;
         if (knt<MXLIST-3) knt++;      knote_mm++;  /* first note flag */
         continue;   /* defer duration until E_END_CHORD seen */
       }  /* end if (inkord) */
     } /* if (tone>=0) determine velocity */

     if (tone<0) { /* various special markers in abc... */
           delta_when = durub = dur = kordur = 0; silent = 1;
	  if (tone==E_START_CHORD) { /* open bracket */ inkord = 1; knt = 0;
	                  silence = 0;  /* zero delay from previous staccato */}
	  else if (tone==E_END_CHORD) { /* end chord bracket */ 
  /* 052.12 get consistent metric duration, 054.05 delta_when ... */
	    delta_when = 
	    dur = durub = kordur = chord_duration(durs,knt); silent = 0; 
          } else if (tone==E_SLUR_START) 
              { if (qslur&1) inslur = 1; knt = 0; when_slur = when;
	}  else if (tone == E_BAR )   /* measure bar seen */
	  {  if (inslur==0) left = sort_gen_events(&kgen,&start_mm,when); 
             if (tictoc_mm>0) meas_mm++;  tictoc_mm = 0; 
 if (Q_VERBOSE>4 && V_CLASS==4) 
  printf(" measure %3d now_met %6d when %6d\n ",meas_mm,now_met,when);
	       if (tictoc_mm>0 && Q_VERBOSE>1 && V_CLASS==4) 
                printf("measure  %3d  notes %4d  duration %4d  nout %5d\n",
			   meas_mm,tictoc_mm/quantum,tictoc_mm,nout);
               if (Q_VERBOSE>2 && V_CLASS==4)
	       printf("last out %d now_mm %d left %d\n",start_mm,now_met,left); 
  	  }
     }  /* (tone<0) */

     /* modify duration if rubato */
       if (qsynch) { /* synchronize metric and rubato time, determine accent */
	 met_frac += dur;  rq = 0;  dq = 0.0;
	 while (met_frac>=quantur) 
           { met_frac -= quantur; 
	     if (qrub) { durub = phrasing[krub]; delay = phrasing[krub+1];
	                 krub += 2; if (krub>=lrub) krub = irub; } 
	     else { durub = quantur - staccato; delay = staccato; }
	     now_step += durub + delay;  
              if (qacc) { kacc++; if (kacc>=lacc) kacc = iacc; } 
           }  /* while (met_frac...) */   
	 if (met_frac>0) 
           {  if (qrub) { durub = phrasing[krub]; delay = phrasing[krub+1];}
	      else { durub = quantur - staccato; delay = staccato; }
	      ddd = durub + delay;  dq = (met_frac*ddd)/quantur;  
              rq = dq + 0.5; }
         delta_when = now_step + rq - when;  
	 if (delay>=0) { silence = delay; durub = delta_when; 
        	 if (2*delay>durub) { delay = durub/2;  silence = delay; } }
         else { /* guitar strum case... think about this */
	         silence = 0; if (qrub) durub *= (float)dur/(float)quantur;
		              else durub = dur; }
       }  /* if (qsynch) */ 
       
       if (chords==2 && nbeat<16 && Q_VERBOSE>1 && V_CLASS==4)
	 printf("nbeat %3d %d %d %d tone %3d vol %3d\n",
		nbeat,down_beat,up_beat,off_beat,tone,vol);

       if (Q_VERBOSE>6 && V_CLASS==4) printf(" %4d kacc  %4d tone %4d vol  %4d dur\n",
				 kacc,tone,vol,kordur);

       if (tone==E_END_CHORD) { /* end chord bracket: NB durs are metric, not rubato*/ 
              make_notes(knt,kord,durs,vels,when,silence,chan,0); 
                 inkord = 0;  knt = 0; 
            when += delta_when;  now_met += dur; 
            tictoc_mm += dur; pedometer(dur); 
       } else if (tone==E_SLUR_END) {
              if (inslur) make_notes(knt,slur,durs,vels,when_slur,silence,chan,qslur); 
                inslur = 0; knt = 0;
       } else if (tone>=0) 
         { /* default melody behavior  (or non chords in vamp 0.35.26) */

	                  if (vol>0) { knote_mm++; last_mm = tone; }

            if (tone<0||tone>127) printf("bad tone %d in melody\n",tone);
          tone += trans;  if (tone<0) tone = 0; if (tone>127) tone = 127;

	  if (inslur>0) { slur[knt] = tone; durs[knt] = durub; vels[knt] = vol;
	                  if (knt<MXLIST-3) knt++; } 
          else { if (pitch[n+1]==E_TIE)  /* handle multiple(?) tie(s) */ 
	          { if (inslur==0) { inslur = -1; knt = 0; when_slur = when; }
		    slur[knt] = tone, durs[knt] = durub; vels[knt] = vol; 
                    if (knt<MXLIST-3) knt++; }
                 else if (inslur<0) { inslur = 0; /* no more ties */
		    slur[knt] = tone, durs[knt] = durub; vels[knt] = vol; 
		   make_notes(knt+1,slur,durs,vels,when_slur,silence,chan,qslur); knt = 0; } 
              else make_notes(1,&tone,&dur,&vol,when,silence,chan,qslur); 
	  } /* (inslur<=0) single notes and tied notes */

       if (Q_VERBOSE>4 && V_CLASS==4) 
             printf(" %3d  tone %4d dur %4d vol %4d\n",n,tone,dur,vol); 
            when += delta_when;  now_met += dur; 
            tictoc_mm += dur; pedometer(dur); 
      } /* melody  ( ! inkord )  */
            gen_epoch = when;   /* 035.17  for event tracking */
 if (Q_VERBOSE>4 && V_CLASS==4) 
 printf(" note %3d now met %d rub %d kordur %d dur %d rub %d\n",
                tone,now_met,when, kordur,dur,durub);
     } /* for (n=npi ... npf) music loop  */
} /* gen_midi_music_track() */

int write_midi_music_track(void)
{  int i;    char i4[4],  *mtrk="MTrk";
/* terminate and write MIDI track */

   if (kgen>0)  /*clean up*/
      sort_gen_events(&kgen,&start_mm,when+4*(max_dither+quantum));

    printf("        pack %d gen %d\n",pack_kount,gen_kount);
      delta(100); shout(0xFF2F); delta(0); /* end of track */

if (fout) {
      fwrite(mtrk,1,4,fout); 
      i = nout;  lowendian(4,(char*)&i,i4);  fwrite(i4,1,4,fout); 
      fwrite(bout,1,nout,fout); 
      printf("MIDI track written: length %d  hex %x  ",nout,nout);
      printf("  ticks %d measures %d\n",gen_epoch,gen_epoch/ticks_in_meas);
      }
return 0;
} /* write_midi_music_track() */

  /* deal with fragments (from repeats) and possible endings 
         "[3" for last time through and (coda != 0)  */
       /* also translate chord list into fragments, 
          vamp should work since count==0   */
  /* 052.16   explicit chunk index for accent, phrase, rubato */

void gen_midi_frags(int kchunk, int chan, int chords, int coda,
                    int acc, int phr, int rub)  
{ int ip, lp, ifrag, lfrag, kount, kf, k, koda=0, type;
  if (kchunk<0 || kchunk>=chunks) 
    { printf("chunk index %d out of range (gen_midi_frags)\n",kchunk);
      return; }
  ip = ifrag = chunk[kchunk].ix; lp = lfrag = chunk[kchunk].lx; 
  kount = chunk[kchunk].count;  /* zero means one fragment [ip:lp-1] */
  type = chunk[kchunk].type;

  /* define flags for gen_midi_music_track()   052.16 */
  /* do not enable for potatoes, only accent for chords */
  qacc = qphr = qrub = qsynch = 0;  quantur = quantum;  when = now_met;
  if (acc> -1) { qacc = 1; kacc = iacc; if (rub> -1) qsynch = 1; }
  if (rub> -1) { if (lrub>irub) qrub = 1; krub = irub;  qsynch = 1;
                 quantur = phrasing[irub-2];  met_frac = 0;
		 when = now_step = now_met + phrasing[irub-1]; }
  if (phr> -1) { if (chords==0 || chords==2) qphr = 1; kphrase = iphrase; }

  if (Q_VERBOSE>2 && (V_CLASS==0 || V_CLASS==4))
    printf("<<<<type %d ifrag %d lfrag %d count %d\n",type,ifrag,lfrag,kount);
  if (kount==0) { gen_midi_music_track(ip,lp,chan,chords); return; }
  if (type == T_MEL) {  /* melody fragments */

  if (kount<(lfrag-ifrag)) /* last fragment is coda [3 ... | */
    { lfrag = lfrag - 2; koda = lfrag; }
     ifrag += 2;  /* 035.09 omit frag[0] (total abc) */
  
  if (coda==0 || koda==0) { for (kf=ifrag; kf<lfrag; kf += 2)
          gen_midi_music_track(frags[kf],frags[kf+1],chan,chords); 
  } else { for (kf=ifrag; kf<lfrag-2; kf += 2)
                gen_midi_music_track(frags[kf],frags[kf+1],chan,chords); 
           gen_midi_music_track(frags[koda],frags[koda+1],chan,chords);  }
  } else if (type == T_CHORD) { /* chord list, no provision for koda */
    for (kf=ifrag; kf<lfrag; kf++)
      { k = frags[kf]; ip = chunk[k].ix; lp = chunk[k].lx;
        if (Q_VERBOSE>4 && (V_CLASS==4)) printf("gen_frag chord %d  i %d f %d\n",k,ip,lp);
            /* sanity check chunk[k].type == T_VAMP ?  */
        gen_midi_music_track(ip,lp,chan,chords); }
  } else if (type == T_VAMP) { printf("count not zero VAMP gen_midi_frags\n");
       gen_midi_music_track(ip,lp,chan,chords); 
  } else printf(" bad type %d (gen_midi_frags)\n",type);
}  /* gen_midi_frags()  */
